// Copyright (2003) Michael L Perry.
// Permission is granted to use this source code without restriction
// privided the copyright notice appears without modification.

// XMLInterpreter.cpp
//
// Uses XMLInterpreter to parse personnel.xml.

#include "PersonnelHandler.h"

IXMLElementHandler *PersonnelHandler::default_child(
   const XMLCh *const name )
{
   // Report an unexpected element.
   std::stringstream err;
   err << "Unexpected element \"" << to_basic_string(name) << "\".";
   throw std::exception( err.str().c_str() );
}

void PersonnelHandler::person_start(
   AttributeList &attrs )
{
   // Start a new person with a new ID.
   m_current = Person();
   const XMLCh *id = attrs.getValue( L"id" );
   if ( id == NULL )
      throw std::exception( "Expecting \"id\" attribute." );
   m_id = to_basic_string( id );
}

void PersonnelHandler::person_end()
{
   // Put the person into the collection.
   m_people[m_id] = m_current;
}

IXMLElementHandler *PersonnelHandler::child(
   const XMLCh *const name )
{
   // Return the facet to handle the root element.
   if ( stringEqual( name, L"personnel" ) )
      return &f_personnel;
   return default_child( name );
}

IXMLElementHandler *PersonnelHandler::personnel_child(
   const XMLCh *const name )
{
   // Return the facet to handle the sub-element.
   if ( stringEqual( name, L"person" ) )
      return &f_person;
   return default_child( name );
}

IXMLElementHandler *PersonnelHandler::person_child(
   const XMLCh *const name )
{
   // Return the facet to handle the sub-element.
   if ( stringEqual( name, L"name" ) )
      return &f_name;
   if ( stringEqual( name, L"email" ) )
      return &f_email;
   if ( stringEqual( name, L"web" ) )
      return &f_web;
   if ( stringEqual( name, L"link" ) )
      return &f_link;
   return default_child( name );
}

IXMLElementHandler *PersonnelHandler::name_child(
   const XMLCh *const name )
{
   // Return the facet to handle the sub-element.
   if ( stringEqual( name, L"family" ) )
      return &f_familyName;
   if ( stringEqual( name, L"given" ) )
      return &f_givenName;
   return default_child( name );
}

IXMLElementHandler *PersonnelHandler::web_child(
   const XMLCh *const name )
{
   // Return the facet to handle the sub-element.
   if ( stringEqual( name, L"work" ) )
      return &f_workWeb;
   if ( stringEqual( name, L"family" ) )
      return &f_familyWeb;
   return default_child( name );
}

void PersonnelHandler::familyName_characters(
   const XMLCh *const chars,
   const unsigned int length)
{
   // Store the family name.
   m_current.familyName = to_basic_string( chars );
}

void PersonnelHandler::givenName_characters(
   const XMLCh *const chars,
   const unsigned int length)
{
   // Store the given name.
   m_current.givenName = to_basic_string( chars );
}

void PersonnelHandler::email_characters(
   const XMLCh *const chars,
   const unsigned int length)
{
   // Store the email address.
   m_current.email = to_basic_string( chars );
}

void PersonnelHandler::workWeb_characters(
   const XMLCh *const chars,
   const unsigned int length)
{
   // Store the work web address.
   m_current.workWeb = to_basic_string( chars );
}

void PersonnelHandler::familyWeb_characters(
   const XMLCh *const chars,
   const unsigned int length)
{
   // Store the family web address.
   m_current.familyWeb = to_basic_string( chars );
}

void PersonnelHandler::link_start(
   AttributeList &attrs )
{
   // Scan for all subordinate IDs.
   const char delims[] = " ";
   const XMLCh *value = attrs.getValue( L"subordinates" );
   if ( value != NULL )
   {
      std::string values = to_basic_string( value );
      size_t token = values.find_first_not_of( delims );
      while ( token != std::string::npos )
      {
         size_t delim = values.find_first_of( delims, token );
         if ( delim == std::string::npos )
         {
            m_current.employees.push_back( values.substr(token) );
            break;
         }
         else
         {
            m_current.employees.push_back( values.substr(token, delim-token) );
         }
         token = values.find_first_not_of( delims, delim );
      }
   }

   // Record the manager ID.
   value = attrs.getValue( L"manager" );
   if ( value != NULL )
      m_current.employer = to_basic_string( value );
}

void PersonnelHandler::parse( const char * const systemID )
{
   XMLInterpreter interpreter( this );

   interpreter.parse( systemID );
}
