// Copyright (2003) Michael L Perry.
// Permission is granted to use this source code without restriction
// privided the copyright notice appears without modification.

// XMLInterpreter.cpp
//
// Defines an interpreter for XML files. Implemented using Xerces for C
// 2.3 SAX parser.

#include "XMLInterpreter.h"

XMLInterpreter::XMLInterpreter( IXMLDocumentHandler *documentHandler ) :
   m_documentHandler( documentHandler ),
   m_locator( NULL )
{
}

void XMLInterpreter::parse( const char *const systemID )
{
   // Create a SAX parser, and parse the document.
   SAXParser parser;
   parser.setDocumentHandler( this );
   parser.parse( systemID );
}

void XMLInterpreter::characters (const XMLCh *const chars, const unsigned int length)
{
   // Notify the top element handler of the characters.
   m_elementHandlers.top()->characters( chars, length );
}

void XMLInterpreter::endDocument ()
{
   // Notify the document handler of the end of the document.
   m_documentHandler->end();
}

void XMLInterpreter::endElement (const XMLCh *const name)
{
   // Notify the top element handler of the end of the element, then
   // pop it, revealing the handler of it's container.
   m_elementHandlers.top()->end();
   m_elementHandlers.pop();
}

void XMLInterpreter::ignorableWhitespace (const XMLCh *const chars, const unsigned int length)
{
   // We ignore whitespace.
}

void XMLInterpreter::processingInstruction (const XMLCh *const target, const XMLCh *const data)
{
   // We ignore processing instructions.
}

void XMLInterpreter::resetDocument ()
{
   // Prepare to reuse the document handler.
   m_documentHandler->reset();
   while ( !m_elementHandlers.empty() )
      m_elementHandlers.pop();
   m_locator = NULL;
}

void XMLInterpreter::setDocumentLocator (const Locator *const locator)
{
   // Store the locator pointer.
   m_locator = locator;
}

void XMLInterpreter::startDocument ()
{
   // Notify the document handler of the start of the document.
   m_documentHandler->start();
}

void XMLInterpreter::startElement (const XMLCh *const name, AttributeList &attrs)
{
   // If we are parsing an element, use its handler.
   // Otherwise, use the documents handler itself.
   IXMLElementContainerHandler *top;
   if ( !m_elementHandlers.empty() )
      top = m_elementHandlers.top();
   else
      top = m_documentHandler;

   // Obtain the element handler of the new element from the current scope.
   IXMLElementHandler *newElement = top->child( name );
   m_elementHandlers.push( newElement );

   // Notify the element handler of the start of the element.
   newElement->start( attrs );
}
