// Copyright (2003) Michael L Perry.
// Permission is granted to use this source code without restriction
// privided the copyright notice appears without modification.

// XMLInterpreter.h
//
// Defines an interpreter for XML files. Implemented using Xerces for C
// 2.3 SAX parser.

#ifndef _XMLINTERPRETER_H
#define _XMLINTERPRETER_H

#include <xercesc/parsers/SAXParser.hpp>
#include <xercesc/sax/DocumentHandler.hpp>
#include <stack>

XERCES_CPP_NAMESPACE_USE

class IXMLElementHandler;

// An interface for handling a container of XML elements.
class IXMLElementContainerHandler
{
protected:
   ~IXMLElementContainerHandler() {}
public:
   // Return a handler for a child element.
   virtual IXMLElementHandler *child(
      const XMLCh *const name ) = 0;
};

// An interface for handling an XML element.
class IXMLElementHandler : public IXMLElementContainerHandler
{
protected:
   ~IXMLElementHandler() {}
public:
   // Handle the start of the element (including attributes).
   virtual void start(
      AttributeList &attrs ) = 0;
   // Handle the end of the element.
   virtual void end() = 0;
   // Handle a chunk of character data within the element.
   virtual void characters(
      const XMLCh *const chars,
      const unsigned int length) = 0;
};

// An interface for handling an XML document.
class IXMLDocumentHandler : public IXMLElementContainerHandler
{
protected:
   ~IXMLDocumentHandler() {}
public:
   // Handle the start of the document.
   virtual void start() = 0;
   // Handle the end of the document.
   virtual void end() = 0;
   // Prepare to start a new document.
   virtual void reset() = 0;
};

class XMLInterpreter : private DocumentHandler
{
private:
   // The document handler.
   IXMLDocumentHandler *m_documentHandler;
   // A stack of element handlers.
   std::stack< IXMLElementHandler * > m_elementHandlers;
   // The locator of the source file.
   const Locator * m_locator;

public:
   XMLInterpreter( IXMLDocumentHandler *documentHandler );

   void parse( const char *const systemID );

private:
   // DocumentHandler implementation.
   void  characters (const XMLCh *const chars, const unsigned int length);
   void  endDocument ();
   void  endElement (const XMLCh *const name);
   void  ignorableWhitespace (const XMLCh *const chars, const unsigned int length);
   void  processingInstruction (const XMLCh *const target, const XMLCh *const data);
   void  resetDocument ();
   void  setDocumentLocator (const Locator *const locator);
   void  startDocument ();
   void  startElement (const XMLCh *const name, AttributeList &attrs);
};

#endif
